unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, DdeMan, StdCtrls;

type
  TMainForm = class(TForm)
    Panel1: TPanel;
    PollTime: TTimer;
    PTData: TDdeServerConv;
    Live: TDdeServerItem;
    DDELabel: TLabel;
    IntervalLabel: TLabel;
    CopyDDEButton: TButton;
    SetIntervalButton: TButton;
    IntervalEdit: TEdit;
    MainPaint: TPaintBox;
    HoverPanel: TPanel;
    MouseOutTime: TTimer;
    LogButton: TButton;
    HoverLabel: TLabel;
    LogLabel: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure PollTimeTimer(Sender: TObject);
    procedure PTDataOpen(Sender: TObject);
    procedure PTDataClose(Sender: TObject);
    procedure CopyDDEButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SetIntervalButtonClick(Sender: TObject);
    procedure MainPaintPaint(Sender: TObject);
    procedure MainPaintMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure MouseOutTimeTimer(Sender: TObject);
    procedure LogButtonClick(Sender: TObject);
  protected
    procedure LogHeaderRow;
    procedure SetDDEConnectionsStr;
  public
    procedure IdleFinalize(Sender: TObject; var Done: Boolean);
    procedure LoadMyGroup;
  end;

var
  MainForm: TMainForm;

procedure LogLine(Lin: String);





implementation

{$R *.DFM}

uses
  Math, AIOUSB, DFileStreams, TheseUtils, ScanBitmaps, LogUnit;

const
  RecoveryStr = 'please hotplug the device and then restart DPM';  
  DisplayRows = 4;
  DisplayCols = 16;

type
  TChData = record
    V, T: Double;
  end;

  T3FColor = record
    R, G, B: Single;
  end;

  TBytes = array of Byte;

var
  MyIndex: LongWord;
  MyConfig: TBytes;
  LogFile: TDFileStream;
  LogHour: Integer;
  DDEItem: TDdeServerItem;
  DDEConnections: LongWord;
  DataForDDE: TStringList;
  ChData: array[0..64-1] of TChData;
  oW, oH: Integer;
  BackBuffer: TScanBitmap;
  NumberFontScale: Single = 2/3;
  HoverChannel: Integer = -1;
  HoverStr: String;

function MyFloatToStr(F: Double): String;
begin
  if IsNAN(F) then
    Result := ''
  else
    Result := FloatToStrF(F, ffFixed, 15, 3)
  ;
end;

procedure LogLine(Lin: String);
var
  I, oI: Integer;
  CtStr: String;

begin
  oI := LogForm.LogList.ItemIndex;
  if oI + 1 = LogForm.LogList.Items.Count then oI := -1;
  I := LogForm.LogList.Items.Add(Lin);
  if oI = -1 then begin
    LogForm.LogList.ItemIndex := I;
  end;

  //CtStr := IntToStr(LogForm.LogList.Items.Count);
  I := LogForm.LogList.Items.Count;
  if I > 64 then begin
    LogForm.LogList.Items.Delete(0);
    CtStr := '64+';
  end
  else begin
    CtStr := IntToStr(I);
  end;

  MainForm.LogButton.Caption := 'View Log(' + CtStr + ')';
end;

procedure CloseOtherDevices;
var
  iDI: LongWord;

begin
  for iDI := 0 to 31 do begin
    if iDI = MyIndex then Continue;
    AIOUSB_CloseDevice(iDI);
  end;
end;

function IDToStr(ID: LongWord): String;
begin
  if (ID and $FFFF0000) <> $0EA60000 then
    Result := 'device #' + IntToHex(ID, 8) + '(?)'
  else begin
    Result := ' of ' + IntToStr((ID shr 8) and $FF);
    case (ID and $FF) of
      0: Result := 'V-First' + Result;
      1: Result := 'V-Last' + Result;
      2: Result := 'Temp' + Result;
      else Result := '#' + IntToStr(ID and $FF) + '(?)' + Result;
    end;
  end;
end;

procedure BuildConfig(var Config: TBytes);
var
  I: Integer;

begin
  SetLength(Config, 21);
  FillChar(Config[0], Length(Config), 0);
  for I := 0 to 7 do Config[I] := $06;
  for I := 8 to 15 do Config[I] := $02;
  Config[$11] := $14; //Software-start scans.
  Config[$13] := 1; //Oversample +1; this is just for damping noise, ADC_GetScan will reject the first one for us.
  //Config[$13] := 0; //Not using ADC_GetScan, so no oversample.

  //This will be overwritten for "split" devices.
  Config[$12] := $F0; //Scan all 128 channels.
  Config[$14] := $70;
end;

procedure ConfigureAI(var DeviceIndex: LongWord; const Config: TBytes);
var
  ConfigSize, Status: LongWord;
  //Hz: Double;

begin
  ////Stop the clock, just in case.
  //Hz := 0;
  //Status := CTR_StartOutputFreq(DeviceIndex, 0, @Hz);

  ConfigSize := Length(Config);
  Status := ADC_SetConfig(DeviceIndex, @Config[0], ConfigSize);
  if Status <> ERROR_SUCCESS then begin
    LogLine('Error ' + IntToStr(Status) + ' configuring A/D; ' + RecoveryStr + '.');
    AIOUSB_CloseDevice(DeviceIndex);
    DeviceIndex := diNone;
    Exit;
  end;

  {
  if ADC_QueryCal(DeviceIndex) = ERROR_SUCCESS then begin
    Status := ADC_SetCal(DeviceIndex, ':AUTO:');
    if Status <> ERROR_SUCCESS then begin
      LogLine('Error ' + IntToStr(Status) + ' calibrating ' + IDToStr(ID) + '; ' + RecoveryStr + '.');
      AIOUSB_CloseDevice(DeviceIndex);
      DeviceIndex := diNone;
      Exit;
    end;
  end;
  }
end;

function BuildLogFN(const ST: TSystemTime): String;
begin
  with ST do Result := 'DPM PT ' + Format('%.4d.%.2d.%.2d %.2d', [ wYear, wMonth, wDay, wHour ]) + '.csv';
end;



procedure TMainForm.FormCreate(Sender: TObject);
var
  I: Integer;
  CmdStr, PrmStr: String;

begin
  DDEItem := Live;

  for I := 0 to High(ChData) do begin
    ChData[I].V := MyNAN;
    ChData[I].T := MyNAN;
  end;

  DataForDDE := TStringList.Create;

  BackBuffer := TScanBitmap.Create;
  BackBuffer.Canvas.Font.Color := $606060;
  BackBuffer.Canvas.Font.Name := 'Tahoma';
  BackBuffer.Canvas.Font.Size := 256;
  I := BackBuffer.Canvas.TextWidth('200');
  NumberFontScale := 7/8 * 256/I;

  BuildConfig(MyConfig);

  for I := 1 to ParamCount do begin
    SplitStr(ParamStr(I), '=', CmdStr, PrmStr);
    if Copy(CmdStr, 1, 1) = '/' then CmdStr := Copy(CmdStr, 2, MAXINT);

    if CmdStr = 'I' then try
      IntervalEdit.Text := Trim(PrmStr);
      SetIntervalButton.Click;
    except
      IntervalEdit.Text := '10';
      LogLine('Error: Invalid interval "' + PrmStr + '" on the command line.');
    end;
  end;

  Application.OnIdle := IdleFinalize;

  SetIntervalButton.Click;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(DataForDDE);
  FreeAndNil(BackBuffer);
end;

procedure TMainForm.IdleFinalize(Sender: TObject; var Done: Boolean);
begin
  Application.OnIdle := nil;

  {
  if MyGroup < 0 then begin
    DetectForm := TDetectForm.Create(Self);
    DetectForm.Show;
  end;
  }

  LoadMyGroup;
end;

procedure TMainForm.LoadMyGroup;
var
  ST: TSystemTime;
  CapStr, LogFN: String;
  bNewLog: Boolean;

begin
  CapStr := 'DPM';
  Caption := CapStr;
  LogForm.Caption := CapStr + ' - Log';

  MyIndex := ResolveDeviceIndex(diFirst);

  CloseOtherDevices;

  if MyIndex <> diNone then ConfigureAI(MyIndex, MyConfig);

  //Start logging.
  GetLocalTime(ST);
  LogHour := ST.wHour;
  LogFN := BuildLogFN(ST);
  try
    bNewLog := not FileExists(LogFN);
    LogFile := TDFileStream.Create(LogFN, dfOpenOrCreate + dfReadWrite);
    if bNewLog then LogHeaderRow else LogFile.Seek(0, soFromEnd);
    LogLabel.Caption := 'Logging to:'#13#10 + IncludeTrailingBackslash(GetCurrentDir) + LogFN;
  except
    on E: Exception do begin
      FreeAndNil(LogFile);
      LogLine('Unable to create/open log file "' + LogFN + '".');
      LogLabel.Caption := 'Unable to create/open log file "' + LogFN + '": ' + E.ClassName + '=' + E.Message;
    end;
  end;

  ddeMgr.AppName := CapStr;
  CopyDDEButton.Enabled := True;
  CopyDDEButton.SetFocus;

  PollTime.Enabled := True;
  PollTimeTimer(PollTime);
end;

procedure TMainForm.LogHeaderRow;
var
  IStr, HeaderRow: String;
  I: Integer;

begin
  HeaderRow := '"Time"';
  for I := 0 to High(ChData) do begin
    IStr := IntToStr(I);
    HeaderRow := HeaderRow + (',"P' + IStr + '","T' + IStr + '"');
  end;
  LogFile.WriteStr(HeaderRow + #13#10);
end;

procedure TMainForm.PollTimeTimer(Sender: TObject);
const
  NumCh = 128;

  function TryAcquire(var DeviceIndex: LongWord; pData: PWord; bSplit: Boolean; const Config: TBytes): LongWord;
  var
    I: Integer;
    ConfigSize: LongWord;
    //BytesLeft, Den, Naui: LongWord;

  begin
    FillChar(pData^, SizeOf(pData^) * NumCh, $0000);

    ConfigSize := Length(Config);

    if not bSplit then begin
      Result := ADC_GetScan(DeviceIndex, pData);
    end
    else begin
      for I := 0 to 3 do begin
        Config[$12] := $F0; //Scan chunk of 32 channels.
        Config[$14] := $10 + $22 * I;
        Result := ADC_SetConfig(DeviceIndex, @Config[0], ConfigSize);
        if Result <> ERROR_SUCCESS then Break;

        //Result := GenericVendorWrite(DeviceIndex, $B6, 0, 0, SizeOf(pData^) * 32, pData);
        //Inc(pData, 32);
        Result := ADC_GetScan(DeviceIndex, pData);
        if Result <> ERROR_SUCCESS then Break;
      end;
    end;

    if Result <> ERROR_SUCCESS then begin
      LogLine('Error ' + IntToStr(Result) + ' reading A/D; reinitializing...');
      AIOUSB_CloseDevice(DeviceIndex);
      GetDevices;
      DeviceIndex := ResolveDeviceIndex(diFirst);
      CloseOtherDevices;
      if DeviceIndex <> diNone then ConfigureAI(DeviceIndex, Config);

      if DeviceIndex <> diNone then
        LogLine('Reinitialized.')
      else
        LogLine('Failed to reinitialize, ' + RecoveryStr + '.')
      ;
    end;
  end;

var
  ADData: array of Word;
  Ct: Word;
  I: Integer;
  Status: LongWord;
  DDEStr, DataForLog, VStr, TStr, LogFN: String;
  ST: TSystemTime;
  Temp: Double;

begin
  Status := ERROR_FILE_NOT_FOUND;
  SetLength(ADData, NumCh);

  for I := 0 to High(ChData) do begin
    ChData[I].V := MyNAN;
    ChData[I].T := MyNAN;
  end;

  GetLocalTime(ST);

  if MyIndex <> diNone then
    Status := TryAcquire(MyIndex, @ADData[0], True, MyConfig)
  ;





  if (MyIndex <> diNone) and (Status = ERROR_SUCCESS) then begin
    for I := 0 to High(ChData) do try
      ChData[I].V := PaFromCt(ADData[64 + I]);
    except
    end;
    for I := 0 to High(ChData) do try
      Ct := ADData[I];
      //Break detect.
      if Ct = $FFFF then Continue;
      if Ct = $0000 then Continue;

      Temp := CFromCt(Ct);
      ChData[I].T := Temp;
    except
    end;
  end;


  if ST.wHour <> LogHour then try //It's a new hour, so fork the logs.
    FreeAndNil(LogFile);
    LogFN := BuildLogFN(ST);
    LogFile := TDFileStream.Create(LogFN, dfOpenOrCreate + dfReadWrite);
    LogHour := ST.wHour;
    LogHeaderRow;
    LogLabel.Caption := 'Logging to:'#13#10 + IncludeTrailingBackslash(GetCurrentDir) + LogFN;
  except
    on E: Exception do begin
      LogLine('Unable to create/open log file "' + LogFN + '".');
      LogLabel.Caption := 'Unable to create/open log file "' + LogFN + '": ' + E.ClassName + '=' + E.Message;
      LogFile := nil;
    end;
  end;

  DataForDDE.Text := 'Ch'#9'Pa'#9'C'#13#10;
  with ST do DataForLog := Format('%.2d:%.2d:%.2d', [ wHour, wMinute, wSecond ]);
  for I := 0 to High(ChData) do with ChData[I] do begin
    VStr := MyFloatToStr(V);
    TStr := MyFloatToStr(T);
    DataForDDE.Add(IntToStr(I) + #9 + VStr + #9 + TStr);
    DataForLog := DataForLog + (',' + VStr + ',' + TStr);
  end;
  DDEStr := DataForDDE.Text;
  DDEItem.Text := DDEStr;
  DDEItem.Lines.Text := DDEStr;
  if LogFile <> nil then LogFile.WriteStr(DataForLog + #13#10);

  MainPaintPaint(MainPaint);
end;

procedure TMainForm.SetDDEConnectionsStr;
begin
  if DDEConnections = 0 then
    DDELabel.Caption := ''
  else
    DDELabel.Caption := 'DDE Links: ' + IntToStr(DDEConnections)
  ;
end;

procedure TMainForm.PTDataOpen(Sender: TObject);
begin
  Inc(DDEConnections);
  SetDDEConnectionsStr;
end;

procedure TMainForm.PTDataClose(Sender: TObject);
begin
  Dec(DDEConnections);
  SetDDEConnectionsStr;
end;

procedure TMainForm.CopyDDEButtonClick(Sender: TObject);
begin
  DDEItem.CopyToClipboard;
end;

procedure TMainForm.SetIntervalButtonClick(Sender: TObject);
var
  Interval: Integer;

begin
  Interval := StrToInt(IntervalEdit.Text);
  if Interval <= 0 then Interval := 1;
  IntervalLabel.Caption := 'Every ' + IntToStr(Interval) + 'sec';
  Interval := 1000 * Interval;
  PollTime.Interval := Interval;
end;

procedure DrawColorBar(Buffer: TScanBitmap; BarRect: TRect; const Color: T3FColor);
const
  Black: T3FColor = (R: $00; G: $00; B: $00);
  White: T3FColor = (R: $FF; G: $FF; B: $FF);

var
  ColColor: array of TBGRFColor;
  X, Y, dX: Integer;
  M, F, nF: Single;
  MixColor: T3FColor;
  pC: PBGRFColor;

begin
  with BarRect do begin
    if Left > Right then Swap(Left, Right);
    if Top > Bottom then Swap(Top, Bottom);
    if Right < 0 then Exit;
    if Left >= Buffer.Width then Exit;
    if Bottom < 0 then Exit;
    if Top >= Buffer.Height then Exit;
    //If the bar is completely off the buffer, then hooray we're done.

    X := Right - Left + 1;
    SetLength(ColColor, X);
    //M := 3/X;
    M := 1.5/X;
    for X := 0 to X-1 do begin
      {
      F := Abs(M * X - 1);
      if F <= 1 then begin
        nF := 1 - F;
        MixColor := White;
      end
      else begin
        nF := F - 1;
        F := 2 - F;
        MixColor := Black;
      end;
      }
      nF := Abs(M * X - 1/2);
      F := 1 - nF;
      MixColor := Black;
      with MixColor do begin
        ColColor[X].Red   := Round(R * nF + Color.R * F);
        ColColor[X].Green := Round(G * nF + Color.G * F);
        ColColor[X].Blue  := Round(B * nF + Color.B * F);
      end;
    end;

    dX := -Left;
    if Left < 0 then Left := 0;
    if Right >= Buffer.Width then Right := Buffer.Width - 1;
    if Top < 0 then Top := 0;
    if Bottom >= Buffer.Height then Bottom := Buffer.Height - 1;

    for Y := Top to Bottom do begin
      pC := Buffer.FindScanPixel(Left, Y);
      for X := Left to Right do begin
        pC^.Color := ColColor[X + dX].Color;
        Inc(pC);
      end;
    end;
  end;
end;

procedure DrawBarsPlate(Buffer: TScanBitmap; X, dX, tY, bY, Num: Integer);
var
  Sz: TSize;
  NumStr: String;

begin
  Buffer.Canvas.Rectangle(X - dX, tY, X + dX, bY);
  if Num >= 0 then begin
    NumStr := IntToStr(Num);
    Sz := BackBuffer.Canvas.TextExtent(NumStr);
    Buffer.Canvas.TextOut(X - (Sz.cx div 2), tY, NumStr);
  end;
end;

procedure TMainForm.MainPaintPaint(Sender: TObject);
const
  TB = 0;
  BaseTM = 1/(2500-TB);
  VoltColor: T3FColor = (R: $C0; G: $40; B: $40);
  TempColor: T3FColor = (R: $40; G: $40; B: $C0);

var
  X, Y, H, W, H2, W2, dW,
  iChannel: Integer;
  VF, TM, F: Double;
  BaseRect, BarRect: TRect;
  NewHoverStr: String;

begin
  W := MainPaint.Width;
  H := MainPaint.Height;
  BackBuffer.Canvas.Brush.Color := $000000;
  if (W = oW) and (H = oH) then begin
    BackBuffer.Canvas.FillRect(Rect(0, 0, W, H));
  end
  else begin
    BackBuffer.Height := 0;
    BackBuffer.Width := W;
    BackBuffer.Height := H;
    BackBuffer.ClearTable;
    oW := W;
    oH := H;
  end;
  BackBuffer.Canvas.Brush.Style := bsClear;
  //Switch W and H from the overall area to each battery.
  W := Max(5, W div DisplayCols); //W is outer, because we don't have a lot of X pixels to spare.
  H := Max(1, (H div DisplayRows) - 8); //H is inner, because Y pixels are our payload.
  H2 := H div 2;
  W2 := W div 2;
  dW := Round(W / 4);

  BackBuffer.Canvas.Pen.Color := $181818;
  X := W * (DisplayCols-1) + W;
  for iChannel := 0 to DisplayRows-1 do begin
    Y := 4 + (H+8) * iChannel + H2;
    BackBuffer.Canvas.MoveTo(0, Y);
    BackBuffer.Canvas.LineTo(X, Y);
  end;
  BackBuffer.Canvas.Pen.Color := $303030;

  BackBuffer.Canvas.Font.Size := Round(NumberFontScale * W);
  VF := 1/2000 * H;
  TM := H * BaseTM;
  for iChannel := 0 to High(ChData) do with ChData[iChannel] do begin
    X := iChannel mod DisplayCols;
    Y := iChannel div DisplayCols;
    with BaseRect do begin
      Left := W * X + W2;
      Right := Left;
      Top := 4 + (H+8) * Y;
      Bottom := Top + H;
    end;

    if iChannel = HoverChannel then begin
      BackBuffer.Canvas.Font.Color := $C0C0C0;
      BackBuffer.Canvas.Pen.Color := $606060;
      DrawBarsPlate(BackBuffer, BaseRect.Left, dW, BaseRect.Top, BaseRect.Bottom, iChannel);
      BackBuffer.Canvas.Font.Color := $606060;
      BackBuffer.Canvas.Pen.Color := $303030;
    end
    else if (iChannel mod 5) = 0 then begin
      DrawBarsPlate(BackBuffer, BaseRect.Left, dW, BaseRect.Top, BaseRect.Bottom, iChannel);
    end
    else begin
      DrawBarsPlate(BackBuffer, BaseRect.Left, dW, BaseRect.Top, BaseRect.Bottom, -1);
    end;

    if not IsNAN(V) then begin
      F := VF * (V + 1000);
      F := Max(0, Min(H, F));
      BarRect.Left := BaseRect.Left - dW;
      BarRect.Right := BaseRect.Right - 1;
      BarRect.Bottom := BaseRect.Bottom;
      BarRect.Top := BarRect.Bottom - Round(F);
      DrawColorBar(BackBuffer, BarRect, VoltColor);
    end;

    if not IsNAN(T) then begin
      F := (T - TB) * TM;
      F := Max(0, Min(H, F));
      BarRect.Left := BaseRect.Left + 1;
      BarRect.Right := BaseRect.Right + dW;
      BarRect.Top := BaseRect.Bottom - Round(F);
      BarRect.Bottom := BaseRect.Bottom;
      DrawColorBar(BackBuffer, BarRect, TempColor);
    end;
  end;

  BitBlt(MainPaint.Canvas.Handle, 0, 0, BackBuffer.Width, BackBuffer.Height, BackBuffer.Canvas.Handle, 0, 0, SRCCOPY);

  if HoverChannel < 0 then
    NewHoverStr := ''
  else with ChData[HoverChannel] do begin
    if IsNAN(V) then begin
      if IsNAN(T) then
        NewHoverStr := '#' + IntToStr(HoverChannel)
      else
        NewHoverStr := '#' + IntToStr(HoverChannel) + ': ' + MyFloatToStr(T) + 'C'
      ;
    end
    else begin
      if IsNAN(T) then
        NewHoverStr := '#' + IntToStr(HoverChannel) + ': ' + MyFloatToStr(V) + 'Pa'
      else
        NewHoverStr := '#' + IntToStr(HoverChannel) + ': ' + MyFloatToStr(V) + 'Pa, ' + MyFloatToStr(T) + 'C'
      ;
    end;
  end;
  if NewHoverStr <> HoverStr then begin
    HoverStr := NewHoverStr;
    HoverLabel.Caption := HoverStr;
  end;
end;

procedure TMainForm.MainPaintMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  NewHoverChannel,
  W, H: Integer;

begin
  MouseOutTime.Enabled := True;

  if (X < 0) or (Y < 0) or (X >= ClientWidth) or (Y >= ClientHeight) then
    NewHoverChannel := -1
  else begin
    //Transcluded.
    W := MainPaint.Width;
    H := MainPaint.Height;
    //Switch W and H from the overall area to each battery.
    W := Max(5, W div DisplayCols); //W is outer, because we don't have a lot of X pixels to spare.
    H := Max(1, (H div DisplayRows) - 8); //H is inner, because Y pixels are our payload.

    X := X div W;
    Y := Y div (H+8);
    if (X >= DisplayCols) or (Y >= DisplayRows) then
      NewHoverChannel := -1
    else
      NewHoverChannel := Y * DisplayCols + X
    ;
  end;

  if NewHoverChannel = HoverChannel then Exit;

  HoverChannel := NewHoverChannel;
  MainPaintPaint(MainPaint);
end;

procedure TMainForm.MouseOutTimeTimer(Sender: TObject);
var
  MusPt: TPoint;

begin
  GetCursorPos(MusPt);
  MusPt := MainPaint.ScreenToClient(MusPt);
  if not PtInRect(MainPaint.ClientRect, MusPt) then begin
    MouseOutTime.Enabled := False;
    HoverChannel := -1;
    MainPaintPaint(MainPaint);
  end;
end;

procedure TMainForm.LogButtonClick(Sender: TObject);
begin
  LogForm.Visible := True;
end;

end.
