unit TheseUtils;

interface

uses
  SysUtils, Graphics;

type
  {$IFDEF ACCURATE_TREAL}
    TReal = Double;
  {$ELSE}
    TReal = Single;
  {$ENDIF}

  TRGBFColor = packed record
    case Boolean of
    True: (
      Color: TColor;
    );
    False: (
      Red: Byte;
      Green: Byte;
      Blue: Byte;
      Flags: Byte;
    );
  end;
  PRGBFColor = ^TRGBFColor;

  TBGRFColor = packed record
    case Boolean of
    True: (
      Color: TColor;
    );
    False: (
      Blue: Byte;
      Green: Byte;
      Red: Byte;
      Flags: Byte;
    );
  end;
  PBGRFColor = ^TBGRFColor;

const
  MyNAN: Double = 0/0;

function IsNAN(V: Double): Boolean;

procedure Swap(var A, D: LongInt); overload;
procedure Swap(var A, D: Pointer); overload;

function InStrC(C: Char; S: String; StartIndex: Integer = 1): LongWord;

function SplitStr(S: String; SplitChar: Char; var Before, After: String): Integer;



function FFromK(K: Double): Double;
function CFromK(K: Double): Double;
{
function KFromR(R: Double): Double;
function RFromCt(Ct: Double): Double;
function VFromCt(Ct: Double): Double;
}
function CFromCt(Ct: Double): Double;
function PaFromCt(Ct: Double): Double;





implementation

function IsNAN(V: Double): Boolean;
type
  PDouble = ^Double;
  
var
  VBits, NANBits: Int64;

begin
  PDouble(@  VBits)^ := V;
  PDouble(@NANBits)^ := MyNAN;
  Result := VBits = NANBits;
end;

procedure Swap(var A, D: LongInt); overload;
asm
  mov ECX, [EAX]
  xchg ECX, [EDX]
  mov [EAX], ECX
end;

procedure Swap(var A, D: Pointer); overload;
asm
  mov ECX, [EAX]
  xchg ECX, [EDX]
  mov [EAX], ECX
end;

function InStrC(C: Char; S: String; StartIndex: Integer = 1): LongWord;
var
  I: Integer;

begin
  for I := StartIndex to Length(S) do
    if S[I] = C then begin
      Result := I;
      Exit;
    end
  ;

  Result := 0;
end;

function SplitStr(S: String; SplitChar: Char; var Before, After: String): Integer;
begin
  Result := InStrC(SplitChar, S);
  if Result = 0 then begin
    Before := S;
    After  := '';
  end
  else begin
    Before := Copy(S, 1, Result - 1);
    After  := Copy(S, Result + 1, MAXINT);
  end;
end;



function FFromK(K: Double): Double;
begin
  Result := 9/5 * K - 459.67;
end;

function CFromK(K: Double): Double;
begin
 Result := K - 273.15;
end;

{
function KFromR(R: Double): Double;
const //Fitted to table in datasheet, not Steinhart-Hart coefficients.
  A = 11.5798774940093;
  B = 2.26165967743856;
  C2 = 0.007058579728405;
  C = 0.000679414298520793;

var
  LR: Double;

begin
  LR := Ln(R);
  //Result := 1 / (A + LR * B + LR*LR*LR * C);
  Result := 10000 / (A + (B + (C2 + C * LR)*LR)*LR);
end;

function RFromCt(Ct: Double): Double;
const
  R1 = 100000; //100k ohm ref resistor.

begin
  //VA/VRef = R2 / (R2 + R1)
  //VRef is 10V, but Ct is from 0-5V range.
  //Result / (R1 + Result) := Ct / $20000;
  Result := R1 * Ct / ($20000 - Ct);
end;

function VFromCt(Ct: Double): Double;
const
  VScale = (1000+49.9)/49.9;

begin
  //0-1V, with a 49.9k/1M ohm voltage divider in front. I kinda wonder why.
  Result := VScale * (1/$10000 * Ct);
end;
}

function CFromCt(Ct: Double): Double;
const //RTDs are pretty simple.
  Freeze = 0.15015;
  dSeventy = (70 - 0) / (0.1907 - Freeze);

begin
  Result := dSeventy * (1/$10000 * Ct - Freeze);
end;

function PaFromCt(Ct: Double): Double;
begin
  Result := 2500/$10000 * Ct - 1000;
end;

end.
